// boundingBox
// Initial code generated by Softimage SDK Wizard
// Executed Sat Feb 5 19:24:03 EST 2011 by Homam
// 
// Tip: To add a command to this plug-in, right-click in the 
// script editor and choose Tools > Add Command.
// 
// Tip: To get help on a callback, highlight the callback name
// (for example, "Init", "Define", or "Execute") and press F1.

function XSILoadPlugin( in_reg )
{
	in_reg.Author = "Homam Bahnassi";
	in_reg.Name = "boundingBox";
	in_reg.email = "hbahnassi@inframez.com";
	in_reg.Major = 1;
	in_reg.Minor = 0;

	in_reg.RegisterCommand("Oriented_BBox", "Oriented_BBox");
	in_reg.RegisterCommand("AxisAligned_BBox", "AxisAligned_BBox");
	in_reg.RegisterMenu(siMenuTbGetPrimitiveID, "Bounding Box", true, false);
	//RegistrationInsertionPoint - do not remove this line

	return true;
}


function XSIUnloadPlugin( in_reg )
{
	var strPluginName;
	strPluginName = in_reg.Name;
	Application.LogMessage(strPluginName + " has been unloaded.",siVerbose);
	return true;
}


// Oriented_BBox plugin defintion
function Oriented_BBox_Init(in_ctxt)
{
	var oCmd;
	oCmd = in_ctxt.Source;
	oCmd.Description = "Generates best fit oriented bounding box";
	oCmd.Tooltip = "Generates best fit oriented bounding box";
	oCmd.ReturnValue = true;

	var oArgs;
	oArgs = oCmd.Arguments;
	return true;
}

function Oriented_BBox_Execute()
{
	Application.LogMessage("Oriented_BBox_Execute called", siVerbose);
	
	// Getting selected points
	var obj_Selection = Application.Selection;
	var ary_ptsPos = new Array();
	if (obj_Selection(0).Type == "pntSubComponent")
	{
		var subComponent = obj_Selection(0).SubComponent;
		var selectedIndices = new VBArray(subComponent.ElementArray).toArray();
		for (var i = 0; i < selectedIndices.length; i++)
			ary_ptsPos[i] = objToWrld(subComponent.Parent3DObject.ActivePrimitive.Geometry.Points(selectedIndices[i]));
	}
	else
	{
		LogMessage("Nothing to do with " + obj_Selection(0).Name + "Type (" + obj_Selection(0).Type + ")", siVerbose);
		return false;
	}
	
	trns_OBBmatrix = compute_OBB(ary_ptsPos);
	
	var obj_cube = ActiveSceneRoot.AddGeometry( "cube", "MeshSurface" );
	obj_cube.length = 1;
	obj_cube.Kinematics.Local.Transform = trns_OBBmatrix;

	return true;
}

// AxisAligned_BBox plugin defintion
function AxisAligned_BBox_Init(in_ctxt) {
	var oCmd;
	oCmd = in_ctxt.Source;
	oCmd.Description = "Generates axis-aligned bounding box";
	oCmd.Tooltip = "Generates axis-aligned bounding box";
	oCmd.ReturnValue = true;

	return true;
}

function AxisAligned_BBox_Execute() {

	Application.LogMessage("AxisAligned_BBox_Execute called", siVerbose);

	// Getting selected points
	var obj_Selection = Application.Selection;
	var ary_ptsPos = new Array();
	if (obj_Selection(0).Type == "pntSubComponent") {
		var subComponent = obj_Selection(0).SubComponent;
		var selectedIndices = new VBArray(subComponent.ElementArray).toArray();
		for (var i = 0; i < selectedIndices.length; i++)
			ary_ptsPos[i] = objToWrld(subComponent.Parent3DObject.ActivePrimitive.Geometry.Points(selectedIndices[i]));
	}
	else {
		LogMessage("Nothing to do with " + obj_Selection(0).Name + "Type (" + obj_Selection(0).Type + ")", siVerbose);
		return false;
	}

	var v3_center = XSIMath.CreateVector3();
	var v3_sides = XSIMath.CreateVector3();
	compute_AABB(ary_ptsPos, v3_center, v3_sides);

	var obj_cube = ActiveSceneRoot.AddGeometry("cube", "MeshSurface");
	var trns_matrix = XSIMath.CreateTransform();
	trns_matrix.copy(obj_cube.Kinematics.Local.Transform);
	obj_cube.length = 1;
	trns_matrix.SetTranslation(v3_center);
	trns_matrix.SetScaling(v3_sides);
	obj_cube.Kinematics.Local.Transform = trns_matrix;

	return true;
}


// Adding sub menu to all commands
function BoundingBox_Init(in_ctxt) {
	var oMenu;
	oMenu = in_ctxt.Source;
	oMenu.AddCommandItem("Oriented_BBox", "Oriented_BBox");
	oMenu.AddCommandItem("AxisAligned_BBox", "AxisAligned_BBox");
	return true;
}


// Math functions

// SUMMARY: Computes the AABB based for point cloud
// PARAMS: ary_ptsPos is the point cloud positions as stored in an array of the 3D vectors.
//			v3_center_Out is the center point of the computed AABB (output argument)
//			v3_sides_Out is the side lengths of the computed AABB (output argument)
// RETURN: VOID

function compute_AABB (ary_ptsPos,v3_center_Out, v3_sides_Out)
{
	var v3_bMin = XSIMath.CreateVector3();
	var v3_bMax = XSIMath.CreateVector3();
	var v3_sides = XSIMath.CreateVector3();
	var v3_center = XSIMath.CreateVector3();
	v3_bMin.Copy(ary_ptsPos[0]);
	v3_bMax.Copy(ary_ptsPos[0]);

	for (i=0; i<ary_ptsPos.length; i++)
	{
		if (ary_ptsPos[i].x < v3_bMin.x)	v3_bMin.x = ary_ptsPos[i].x;
		if (ary_ptsPos[i].y < v3_bMin.y)	v3_bMin.y = ary_ptsPos[i].y;
		if (ary_ptsPos[i].z < v3_bMin.z)	v3_bMin.z = ary_ptsPos[i].z;
		
		if (ary_ptsPos[i].x > v3_bMax.x)	v3_bMax.x = ary_ptsPos[i].x;
		if (ary_ptsPos[i].y > v3_bMax.y)	v3_bMax.y = ary_ptsPos[i].y;
		if (ary_ptsPos[i].z > v3_bMax.z)	v3_bMax.z = ary_ptsPos[i].z;
	}
	
	v3_sides_Out.x = v3_bMax.x - v3_bMin.x;
	v3_sides_Out.y = v3_bMax.y - v3_bMin.y;
	v3_sides_Out.z = v3_bMax.z - v3_bMin.z;
	
	v3_center_Out.x = v3_sides_Out.x*0.5 + v3_bMin.x;
	v3_center_Out.y = v3_sides_Out.y*0.5 + v3_bMin.y;
	v3_center_Out.z = v3_sides_Out.z*0.5 + v3_bMin.z;
	
	// Debug!
	LogMessage("AABB Center Point -> X: " + v3_center_Out.x + ", Y: " + v3_center_Out.y + ", Z: " + v3_center_Out.z, siVerbose);
	LogMessage("AABB sides -> X: " + v3_sides_Out.x + ", Y: " + v3_sides_Out.y + ", Z: " + v3_sides_Out.z, siVerbose);
}

// SUMMARY: Computes the tight OBB based on brute force algorithm
// PARAMS: ary_ptsPos is the point cloud positions as stored in an array of the 3D vectors.
// RETURN: Transformation matrix for the tight OBB.

function compute_OBB(ary_ptsPos)
{
	var bestVolume = 9e12;
	var v3_center = XSIMath.CreateVector3();
	var v3_sides = XSIMath.CreateVector3();
	
	var ax = 0;
	var ay = 0;
	var az = 0;
	var sweep = 45;
	var steps = 7;
	
	var v3_angle = XSIMath.CreateVector3();
	var trns_matrix = XSIMath.CreateTransform();
	
	var v3_sidesFinal = XSIMath.CreateVector3();
	var trns_matrixFinal = XSIMath.CreateTransform();
	
	compute_AABB (ary_ptsPos, v3_center, v3_sides);
	
	while (sweep >= 1)
	{
		var found = false;
		var stepSize = sweep/steps;
		
		for (var x=ax-sweep; x<=ax+sweep; x+=stepSize)
		{
			for (var y=ay-sweep; y<=ay+sweep; y+=stepSize)
			{
				for (var z=az-sweep; z<=az+sweep; z+=stepSize)
				{
					trns_matrix.SetIdentity();
					trns_matrix.SetRotationFromXYZAnglesValues(
								XSIMath.DegreesToRadians( x ),
								XSIMath.DegreesToRadians( y ),
								XSIMath.DegreesToRadians( z ));
					
					trns_matrix.SetTranslationFromValues(v3_center.x,v3_center.y,v3_center.z);
										
					var v3_sidesTrans = XSIMath.CreateVector3();
					
					transform_OBB(ary_ptsPos, trns_matrix, v3_sidesTrans);
					
					var volume = v3_sidesTrans.x*v3_sidesTrans.y*v3_sidesTrans.z; // the volume of the transformed cube
					
					if ( volume < bestVolume )
					{
						bestVolume = volume;

						v3_sidesFinal.Set(v3_sidesTrans.x,v3_sidesTrans.y,v3_sidesTrans.z);
						v3_angle.Set(ax,ay,az);
						trns_matrixFinal.Copy(trns_matrix);
						found = true;
					}
				}
			}
		}
		
		if (found)
		{
			 ax = v3_angle.x;
			 ay = v3_angle.y;
			 az = v3_angle.z;
			 
			 sweep *= 0.5;
		}
		else
			break;
	}
	
	LogMessage("Center Transition -> X: " + trns_matrixFinal.PosX + ", Y: " + trns_matrixFinal.PosY + ", Z: " + trns_matrixFinal.PosZ, siVerbose);
	LogMessage("Center Rotation -> X: " + trns_matrixFinal.RotX + ", Y: " + trns_matrixFinal.RotY + ", Z: " + trns_matrixFinal.RotZ, siVerbose);
	LogMessage("OBB sides local vector -> X: " + v3_sidesFinal.x + ", Y: " + v3_sidesFinal.y + ", Z: " + v3_sidesFinal.z, siVerbose);
	
	trns_matrixFinal.SetScaling(v3_sidesFinal);
	return trns_matrixFinal;
}

// SUMMARY: Computes the tight OBB based on transoformation matrix
// PARAMS: ary_ptsPos is the point cloud positions as stored in an array of the 3D vectors.
//			trns_matrixIO is the transformation matrix for the OBB (this is updated for output)
//			v3_sides_IO is the side lengths for the OBB stored in 3D vector (this is updated for output)
// RETURN: VOID

function transform_OBB(ary_ptsPos, trns_matrixIO, v3_sides_IO)
{
	var m4_invTrans = XSIMath.CreateMatrix4();
	var v3_tmp = XSIMath.CreateVector3();
	var v3_center = XSIMath.CreateVector3();
	var v3_oTranslation = XSIMath.CreateVector3();
	var v3_oCenter = XSIMath.CreateVector3();
	
	var v3_bMax = XSIMath.CreateVector3(-9e12,-9e12,-9e12);
	var v3_bMin = XSIMath.CreateVector3(9e12,9e12,9e12);
	
	m4_invTrans.Invert (trns_matrixIO.Matrix4);
	
	for (i=0; i<ary_ptsPos.length; i++)
	{
		v3_tmp.MulByMatrix4(ary_ptsPos[i], m4_invTrans);

		if (v3_tmp.x < v3_bMin.x)	v3_bMin.x = v3_tmp.x;
		if (v3_tmp.y < v3_bMin.y)	v3_bMin.y = v3_tmp.y;
		if (v3_tmp.z < v3_bMin.z)	v3_bMin.z = v3_tmp.z;
		
		if (v3_tmp.x > v3_bMax.x)	v3_bMax.x = v3_tmp.x;
		if (v3_tmp.y > v3_bMax.y)	v3_bMax.y = v3_tmp.y;
		if (v3_tmp.z > v3_bMax.z)	v3_bMax.z = v3_tmp.z;
	}
	
	v3_center.x = (v3_bMax.x - v3_bMin.x)*0.5 + v3_bMin.x;
	v3_center.y = (v3_bMax.y - v3_bMin.y)*0.5 + v3_bMin.y;
	v3_center.z = (v3_bMax.z - v3_bMin.z)*0.5 + v3_bMin.z;
		
	v3_sides_IO.x = v3_bMax.x - v3_bMin.x;
	v3_sides_IO.y = v3_bMax.y - v3_bMin.y;
	v3_sides_IO.z = v3_bMax.z - v3_bMin.z;
	
	trns_matrixIO.GetTranslation(v3_oTranslation);
	trns_matrixIO.SetTranslationFromValues(0,0,0);
	v3_oCenter.MulByTransformation(v3_center, trns_matrixIO);
	v3_oTranslation.AddInPlace(v3_oCenter);
	trns_matrixIO.SetTranslation(v3_oTranslation);
}

// SUMMARY: Converts point position from object space to global space.
// PARAMS: obj_pt is an XSIObject for point.
// RETURN: Returns a 3D vector of the point's position in global space.
function objToWrld (obj_pt)
{
	// Getting the global transformation matrix for the parent object of the point
	obj_elem = obj_pt.parent.parent.parent;
	trn_elem = obj_elem.kinematics.global.Transform;
	m4_elem = XSIMath.CreateMatrix4();
	trn_elem.GetMatrix4(m4_elem);
	
	v3_pos = obj_pt.position;
	v3_pos.MulByMatrix4InPlace( m4_elem );
	return v3_pos;
}